use std::{cell::RefCell, collections::HashMap};

// use aes_gcm::aes::cipher::BlockEncrypt;
use aes_gcm::{Aes128Gcm, aead::{KeyInit, Aead}, Nonce};
pub use ed25519_dalek::{Keypair, PublicKey, Signature, Signer};

use crate::prelude::*;

pub enum TrustLevel {
    Unknown,
    Distrusted,
    // Future extension
    // Transitively,
    Ultimately,
}

pub struct TrustDB {}

#[derive(serde::Serialize, serde::Deserialize, Default, Debug)]
pub enum PacketHeader {
    #[default]
    NoCrypto,
    Signed(Signature),
    // u128 here is the nonce
    Encrypted(u128),
}

pub struct SymmetricKey {
    key: [u8; 16],
    nonce: u128,
}

pub(crate) struct VerifyPacketSignature {
    api: std::rc::Rc<crate::API>,
}

impl Service for VerifyPacketSignature {
    fn new(api: std::rc::Rc<crate::API>) -> Self {
        Self { api }
    }
    fn setup(self: &std::rc::Rc<Self>) {
        self.api
            .channel::<fleck_core::ReceivePacketChannel>()
            .sub_opt(fleck_core::ReceiveOrder::Verify, self, Self::check);
    }
}

impl VerifyPacketSignature {
    fn check(&self, mut message: fleck_core::msg::Message) -> Option<fleck_core::msg::Message> {
        // we don't care about the message if it isn't supposed to be signed
        if !message.saved_params.unwrap().signed {
            return Some(message);
        }

        if let PacketHeader::Signed(signature) = message.crypto_header {
            let pubkey = message.node.as_ref()?.pubkey();
            // reset the crypto header for verification
            message.crypto_header = PacketHeader::NoCrypto;
            let to_verify =
                bincode::serialize(&message).expect("assume that we can re-serialize the message");

            use ed25519_dalek::Verifier;
            pubkey.verify(&to_verify, &signature).ok()?;
            // verification passed!
            Some(message)
        } else {
            // it was supposed to have a signature, but it didn't...
            None
        }
    }
}

pub(crate) struct SignPacket {
    api: std::rc::Rc<crate::API>,
}

impl Service for SignPacket {
    fn new(api: std::rc::Rc<crate::API>) -> Self {
        Self { api }
    }
    fn setup(self: &std::rc::Rc<Self>) {
        self.api.channel::<fleck_core::SendPacketChannel>().sub_ref(
            fleck_core::SendOrder::Sign,
            self,
            Self::sign,
        );
    }
}

impl SignPacket {
    fn sign(&self, message: &mut fleck_core::msg::Message) {
        let params = message
            .saved_params
            .expect("packet got to signing stage without saved parameters?");

        // don't do anything if the packet doesn't want a signature
        if !params.signed {
            return;
        }

        self.api.with_service(|ns: &fleck_core::NodeService| {
            let ns = ns.self_node();
            let keypair = ns.keypair().unwrap();
            let sign_data = bincode::serialize(&message.content).unwrap();
            message.crypto_header = PacketHeader::Signed(keypair.sign(&sign_data));
        });
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug)]
pub struct KeyExchange {
    pubkey: ed25519_dalek::PublicKey,
    key: x25519_dalek::PublicKey,
}

impl fleck_core::msg::MessageParams for KeyExchange {
    const NAME: &'static str = "KeyExchange";
    const ENCRYPTED: bool = false;
    const SIGNED: bool = true;
}

pub enum KeyExchangeState {
    NoState,
    Given(x25519_dalek::PublicKey),
    WaitForResponse(x25519_dalek::EphemeralSecret),
    Completed(x25519_dalek::PublicKey, SymmetricKey),
}

impl std::fmt::Debug for KeyExchangeState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::NoState => f.write_str("KeyExchangeState::NoState"),
            Self::Given(pkey) => f.write_fmt(format_args!(
                "KeyExchangeState::Given {{ pkey: {:?} }}",
                pkey
            )),
            Self::WaitForResponse(_) => {
                f.write_str("KeyExchangeState::WaitForResponse { redacted }")
            },
            Self::Completed(pkey, _) => f.write_fmt(format_args!(
                "KeyExchangeState::Completed {{ pkey: {:?}, redacted }}",
                pkey
            )),
        }
    }
}

pub struct EncryptionService {
    api: std::rc::Rc<crate::API>,
    state: RefCell<HashMap<fleck_core::peer::Peer, KeyExchangeState>>,
}

impl Service for EncryptionService {
    fn new(api: std::rc::Rc<crate::API>) -> Self {
        Self {
            api,
            state: Default::default(),
        }
    }

    fn setup(self: &std::rc::Rc<Self>) {
        self.api.channel::<fleck_core::SendPacketChannel>()
            .sub_opt(fleck_core::SendOrder::Encrypt, self, Self::encrypt);
        self.api.channel::<fleck_core::ReceivePacketChannel>()
            .sub_opt(fleck_core::ReceiveOrder::Decrypt, self, Self::decrypt);

        self.api.with_service(|msg: &fleck_core::MessageService| {
            msg.add_message_type::<KeyExchange>();
        });
        self.api
            .channel::<fleck_core::node::NodeRegistrationChannel>()
            .sub_ref(self, Self::new_node);
        self.api
            .channel::<fleck_core::msg::MessageChannel<KeyExchange>>()
            .sub_eat(self, Self::incoming);
    }
}

impl EncryptionService {
    fn encrypt(&self, mut packet: fleck_core::Message) -> Option<fleck_core::Message> {
        if !packet.saved_params.unwrap().encrypted {
            return Some(packet)
        }

        let mut state = self.state.borrow_mut();
        let mut keyexchange = state.get_mut(packet.peer.as_ref()?)?;
        if let KeyExchangeState::Completed(_, symkey) = &mut keyexchange {
            let cipher = Aes128Gcm::new_from_slice(&symkey.key).expect("key length mismatch?");
            let nonce_bytes : [u8; 12] = u128::to_be_bytes(symkey.nonce)[0..12].try_into().unwrap();
            let nonce : Nonce<_> = nonce_bytes.into();

            // serialize+encrypt with the message type so we don't leak that
            let plaintext = bincode::serialize(&packet.content).ok()?;
            packet.content.ty = 0;
            packet.content.data = cipher.encrypt(&nonce, plaintext.as_slice()).ok()?;

            packet.crypto_header = PacketHeader::Encrypted(symkey.nonce);

            // advance the nonce
            symkey.nonce += 1;

            Some(packet)
        }
        else {
            log::info!("Tried to send packet before encryption handshake was finished");
            None
        }
    }

    fn decrypt(&self, mut packet: fleck_core::Message) -> Option<fleck_core::Message> {
        // no peer = pass along
        if packet.peer.is_none() {
            return Some(packet)
        }

        // check for an encryption header
        let nonce = match packet.crypto_header {
            PacketHeader::Encrypted(nonce) => nonce,
            _ => return Some(packet),
        };

        let mut state = self.state.borrow_mut();

        // we don't have saved_params or anything yet, or even know what the message type is, so
        // we'll just have to trust that if we have an encryption key and there's a crypto header,
        // we should decrypt it.

        // There shouldn't be anything wrong with decrypting a message that claims to be encrypted
        // and is authenticated, right?... TODO: double-check this assertion.
        if let KeyExchangeState::Completed(_, symkey) = state.get_mut(packet.peer.as_ref()?)? {
            let cipher = Aes128Gcm::new_from_slice(&symkey.key).expect("key length mismatch?");
            let nonce_bytes : [u8; 12] = u128::to_be_bytes(nonce)[0..12].try_into().unwrap();
            let nonce : Nonce<_> = nonce_bytes.into();

            let plaintext = cipher.decrypt(&nonce, packet.content.data.as_slice()).ok()?;
            packet.content = bincode::deserialize(&plaintext).ok()?;

            packet.crypto_header = PacketHeader::NoCrypto;

            Some(packet)
        }
        else {
            None
        }
    }

    fn new_node(&self, node: &mut std::rc::Rc<fleck_core::node::Node>) {
        // we may want to initiate a handshake, if we're not already doing that...
        if let Some(peer) = node.peer().as_ref() {
            if !self.state.borrow().contains_key(peer) {
                self.begin_handshake(peer)
            }
        }
    }

    fn begin_handshake(&self, with: &fleck_core::peer::Peer) {
        use rand::SeedableRng;
        let csprng = rand::rngs::StdRng::from_entropy();
        let secret = x25519_dalek::EphemeralSecret::new(csprng);

        // send notification with new secret
        let send = || {
            log::trace!("sending KeyExchange message");
            self.api.queue::<fleck_core::SendPacketChannel>(
                fleck_core::msg::Message::build(KeyExchange {
                    pubkey: self.api.with_service(|ns: &fleck_core::NodeService| {
                        *ns.self_node().pubkey()
                    }),
                    key: x25519_dalek::PublicKey::from(&secret),
                })
                .with_peer(with.clone()),
            );
        };

        let mut states = self.state.borrow_mut();
        let state = states.remove(with).unwrap_or(KeyExchangeState::NoState);
        match state {
            KeyExchangeState::NoState => {
                send();
                states.insert(with.clone(), KeyExchangeState::WaitForResponse(secret));
            },
            KeyExchangeState::Given(pubkey) => {
                send();
                let result = secret.diffie_hellman(&pubkey);

                let sk = SymmetricKey {
                    key: result.as_bytes()[0..16].try_into().unwrap(),
                    nonce: u128::from_be_bytes(result.as_bytes()[16..32].try_into().unwrap()),
                };

                log::trace!("Completed key negotiation with peer!");
                states.insert(with.clone(), KeyExchangeState::Completed(pubkey, sk));
            },
            KeyExchangeState::WaitForResponse(secret) => {
                send();
                states.insert(with.clone(), KeyExchangeState::WaitForResponse(secret));
            },
            KeyExchangeState::Completed(_pubkey, _symmetric_key) => {
                log::warn!("asked to begin handshake when already have a good encryption key --- did the peer die?");
            },
        }
    }

    fn incoming(&self, msg: (fleck_core::msg::Metadata, KeyExchange)) {
        let mut states = self.state.borrow_mut();
        let peer = msg
            .0
            .peer
            .expect("incoming KeyExchange should have a peer?");

        let state = states.remove(&peer).unwrap_or(KeyExchangeState::NoState);
        log::trace!(
            "incoming key exchange message: {:?}, state: {:?}",
            msg.1,
            &state
        );

        // tell the node service about this peer, just in case they don't know already.
        self.api.with_service(|ns: &fleck_core::NodeService| {
            ns.inform_of(msg.1.pubkey, peer.clone());
        });

        match state {
            KeyExchangeState::NoState => {
                states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));

                drop(states);
                self.begin_handshake(&peer);
            },
            KeyExchangeState::Given(peer_pubkey) => {
                if peer_pubkey != msg.1.key {
                    log::info!("peer DH key changed!");
                    states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));
                }

                drop(states);
                self.begin_handshake(&peer);
            },
            KeyExchangeState::WaitForResponse(secret) => {
                // key handshake is finished!
                let result = secret.diffie_hellman(&msg.1.key);

                let sk = SymmetricKey {
                    key: result.as_bytes()[0..16].try_into().unwrap(),
                    nonce: u128::from_be_bytes(result.as_bytes()[16..32].try_into().unwrap()),
                };

                states.insert(peer.clone(), KeyExchangeState::Completed(msg.1.key, sk));
                drop(states);
            },
            KeyExchangeState::Completed(peer_pubkey, _symmetric_key) => {
                if peer_pubkey == msg.1.key {
                    // it's a repeat, so nothing to do
                    return;
                }
                // it's a new pubkey, so we want to restart the handshake
                states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));
                drop(states);
                self.begin_handshake(&peer);
            },
        }
        log::trace!(
            "after processing incoming key exchange message, state: {:?}",
            self.state.borrow().get(&peer)
        );
    }
}
